@@ -1,21 +1,21 @@
 #' Retrieve an attribute.
 #'
-#' \code{xml_attrs()} retrieves all attributes values as a named character
-#' vector, \code{xml_attrs() <-} or \code{xml_set_attrs()} sets all attribute
-#' values. \code{xml_attr()} retrieves the value of single attribute and
-#' \code{xml_attr() <-} or \code{xml_set_attr()} modifies its value. If the
-#' attribute doesn't exist, it will return \code{default}, which defaults to
-#' \code{NA}. \code{xml_has_attr()} tests if an attribute is present.
+#' `xml_attrs()` retrieves all attributes values as a named character
+#' vector, `xml_attrs() <-` or `xml_set_attrs()` sets all attribute
+#' values. `xml_attr()` retrieves the value of single attribute and
+#' `xml_attr() <-` or `xml_set_attr()` modifies its value. If the
+#' attribute doesn't exist, it will return `default`, which defaults to
+#' `NA`. `xml_has_attr()` tests if an attribute is present.
 #'
 #' @inheritParams xml_name
 #' @param attr Name of attribute to extract.
 #' @param default Default value to use when attribute is not present.
-#' @return \code{xml_attr()} returns a character vector. \code{NA} is used
+#' @return `xml_attr()` returns a character vector. `NA` is used
 #'  to represent of attributes that aren't defined.
 #'
-#'  \code{xml_has_attr()} returns a logical vector.
+#'  `xml_has_attr()` returns a logical vector.
 #'
-#'  \code{xml_attrs()} returns a named character vector if \code{x} x is single
+#'  `xml_attrs()` returns a named character vector if `x` x is single
 #'  node, or a list of character vectors if given a nodeset
 #' @export
 #' @examples
@@ -60,25 +60,7 @@
 #' xml_attrs(doc) <- c("b:id" = "one", "f:id" = "two", "id" = "three")
 #' xml_set_attrs(doc, c("b:id" = "one", "f:id" = "two", "id" = "three"))
 xml_attr <- function(x, attr, ns = character(), default = NA_character_) {
-  UseMethod("xml_attr")
-}
-
-#' @export
-xml_attr.xml_missing <- function(x, attr, ns = character(), default = NA_character_) {
-  default
-}
-
-#' @export
-xml_attr.xml_node <- function(x, attr, ns = character(),
-                              default = NA_character_) {
-  node_attr(x$node, name = attr, missing = default, nsMap = ns)
-}
-
-#' @export
-xml_attr.xml_nodeset <- function(x, attr,  ns = character(),
-                                 default = NA_character_) {
-  vapply(x, xml_attr, attr = attr, default = default, ns = ns,
-    FUN.VALUE = character(1))
+  .Call(node_attr, x, attr, as.character(default), ns)
 }
 
 #' @export
@@ -90,22 +72,7 @@ xml_has_attr <- function(x, attr, ns = character()) {
 #' @export
 #' @rdname xml_attr
 xml_attrs <- function(x, ns = character()) {
-  UseMethod("xml_attrs")
-}
-
-#' @export
-xml_attrs.xml_missing <- function(x, ns = character()) {
-  NA_character_
-}
-
-#' @export
-xml_attrs.xml_node <- function(x, ns = character()) {
-  node_attrs(x$node, nsMap = ns)
-}
-
-#' @export
-xml_attrs.xml_nodeset <- function(x, ns = character()) {
-  lapply(x, xml_attrs, ns = ns)
+  .Call(node_attrs, x, nsMap = ns)
 }
 
 #' @param value character vector of new value.
@@ -118,17 +85,25 @@ xml_attrs.xml_nodeset <- function(x, ns = character()) {
 #' @export
 `xml_attr<-.xml_node` <- function(x, attr, ns = character(), value) {
   if (is.null(value)) {
-    node_remove_attr(x$node, name = attr, nsMap = ns)
+    .Call(node_remove_attr, x$node, attr, ns)
   } else {
     value <- as.character(value)
-    node_set_attr(x$node, name = attr, nsMap = ns, value)
+    .Call(node_set_attr, x$node, attr, value, ns)
   }
   x
 }
 
 #' @export
 `xml_attr<-.xml_nodeset` <- function(x, attr, ns = character(), value) {
-  lapply(x, `xml_attr<-`, attr = attr, ns = ns, value = value)
+  if (length(x) == 0) {
+    return(x)
+  }
+
+  if (length(value) == 0) {
+    value <- list(value)
+  }
+
+  mapply(`xml_attr<-`, x, attr = attr, value = value, SIMPLIFY = FALSE, MoreArgs = list(ns = ns))
   x
 }
 
@@ -167,7 +142,7 @@ xml_set_attr.xml_missing <- set_attr_fun
 #' @export
 `xml_attrs<-.xml_node` <- function(x, ns = character(), value) {
   if (!is_named(value)) {
-    stop("`value` must be a named character vector or `NULL`", call. = FALSE)
+    cli::cli_abort("{.arg value} must be a named character vector or `NULL`")
   }
 
   attrs <- names(value)
@@ -183,7 +158,7 @@ xml_set_attr.xml_missing <- set_attr_fun
 
   # replace existing attributes and add new ones
   Map(function(attr, val) {
-      xml_attr(x, attr, ns) <- val
+    xml_attr(x, attr, ns) <- val
   }, attr = c(existing, new), value[c(existing, new)])
 
 
@@ -201,13 +176,13 @@ xml_set_attr.xml_missing <- set_attr_fun
     return(x)
   }
   if (!is.list(ns)) {
-     ns <- list(ns)
+    ns <- list(ns)
   }
   if (!is.list(value)) {
-     value <- list(value)
+    value <- list(value)
   }
   if (!all(vapply(value, is_named, logical(1)))) {
-    stop("`value` must be a list of named character vectors")
+    cli::cli_abort("{.arg {value}} must be a list of named character vectors.")
   }
 
   Map(`xml_attrs<-`, x, ns, value)
