@@ -10,11 +10,13 @@
 #' `xml_find_one()` has been deprecated. Instead use
 #' `xml_find_first()`.
 
-#' @param xpath A string containing a xpath (1.0) expression.
+#' @param xpath A string containing an xpath (1.0) expression.
 #' @inheritParams xml_name
-#' @return `xml_find_all` always returns a nodeset: if there are no matches
-#'   the nodeset will be empty. The result will always be unique; repeated
-#'   nodes are automatically de-duplicated.
+#' @param ... Further arguments passed to or from other methods.
+#' @return `xml_find_all` returns a nodeset if applied to a node, and a nodeset
+#'   or a list of nodesets if applied to a nodeset. If there are no matches,
+#'   the nodeset(s) will be empty. Within each nodeset, the result will always
+#'   be unique; repeated nodes are automatically de-duplicated.
 #'
 #'   `xml_find_first` returns a node if applied to a node, and a nodeset
 #'   if applied to a nodeset. The output is *always* the same size as
@@ -46,11 +48,16 @@
 #' </body>")
 #' para <- xml_find_all(x, ".//p")
 #'
-#' # If you apply xml_find_all to a nodeset, it finds all matches,
-#' # de-duplicates them, and returns as a single list. This means you
+#' # By default, if you apply xml_find_all to a nodeset, it finds all matches,
+#' # de-duplicates them, and returns as a single nodeset. This means you
 #' # never know how many results you'll get
 #' xml_find_all(para, ".//b")
 #'
+#' # If you set flatten to FALSE, though, xml_find_all will return a list of
+#' # nodesets, where each nodeset contains the matches for the corresponding
+#' # node in the original nodeset.
+#' xml_find_all(para, ".//b", flatten = FALSE)
+#'
 #' # xml_find_first only returns the first match per input node. If there are 0
 #' # matches it will return a missing node
 #' xml_find_first(para, ".//b")
@@ -67,31 +74,38 @@
 #' ')
 #' xml_find_all(x, ".//f:doc")
 #' xml_find_all(x, ".//f:doc", xml_ns(x))
-xml_find_all <- function(x, xpath, ns = xml_ns(x)) {
+xml_find_all <- function(x, xpath, ns = xml_ns(x), ...) {
   UseMethod("xml_find_all")
 }
 
 #' @export
-xml_find_all.xml_missing <- function(x, xpath, ns = xml_ns(x)) {
+xml_find_all.xml_missing <- function(x, xpath, ns = xml_ns(x), ...) {
   xml_nodeset()
 }
 
 #' @export
-xml_find_all.xml_node <- function(x, xpath, ns = xml_ns(x)) {
+xml_find_all.xml_node <- function(x, xpath, ns = xml_ns(x), ...) {
   nodes <- .Call(xpath_search, x$node, x$doc, xpath, ns, Inf)
   xml_nodeset(nodes)
 }
 
+#' @param flatten A logical indicating whether to return a single, flattened
+#'   nodeset or a list of nodesets.
 #' @export
-xml_find_all.xml_nodeset <- function(x, xpath, ns = xml_ns(x)) {
-  if (length(x) == 0)
+#' @rdname xml_find_all
+xml_find_all.xml_nodeset <- function(x, xpath, ns = xml_ns(x), flatten = TRUE, ...) {
+  if (length(x) == 0) {
     return(xml_nodeset())
+  }
 
-  nodes <- unlist(recursive = FALSE,
-    lapply(x, function(x)
-      .Call(xpath_search, x$node, x$doc, xpath, ns, Inf)))
+  res <- lapply(x, function(x) .Call(xpath_search, x$node, x$doc, xpath, ns, Inf))
 
-  xml_nodeset(nodes)
+  if (isTRUE(flatten)) {
+    return(xml_nodeset(unlist(recursive = FALSE, res)))
+  }
+
+  res[] <- lapply(res, xml_nodeset)
+  res
 }
 
 #' @export
@@ -100,6 +114,7 @@ xml_find_first <- function(x, xpath, ns = xml_ns(x)) {
   UseMethod("xml_find_first")
 }
 
+#' @export
 xml_find_first.xml_missing <- function(x, xpath, ns = xml_ns(x)) {
   xml_missing()
 }
@@ -108,7 +123,7 @@ xml_find_first.xml_missing <- function(x, xpath, ns = xml_ns(x)) {
 xml_find_first.xml_node <- function(x, xpath, ns = xml_ns(x)) {
   res <- .Call(xpath_search, x$node, x$doc, xpath, ns, 1)
   if (length(res) == 1) {
-     res[[1]]
+    res[[1]]
   } else {
     res
   }
@@ -116,11 +131,19 @@ xml_find_first.xml_node <- function(x, xpath, ns = xml_ns(x)) {
 
 #' @export
 xml_find_first.xml_nodeset <- function(x, xpath, ns = xml_ns(x)) {
-  if (length(x) == 0)
+  if (length(x) == 0) {
     return(xml_nodeset())
+  }
 
-  xml_nodeset(lapply(x, function(x)
-      xml_find_first(x, xpath = xpath, ns = ns)), deduplicate = FALSE)
+  xml_nodeset(
+    lapply(
+      x,
+      function(x) {
+        xml_find_first(x, xpath = xpath, ns = ns)
+      }
+    ),
+    deduplicate = FALSE
+  )
 }
 
 
@@ -133,23 +156,53 @@ xml_find_num <- function(x, xpath, ns = xml_ns(x)) {
 #' @export
 xml_find_num.xml_node <- function(x, xpath, ns = xml_ns(x)) {
   res <- .Call(xpath_search, x$node, x$doc, xpath, ns, Inf)
-  if (!is.numeric(res)) {
-    stop("result of type: ", sQuote(class(res)), ", not numeric", call. = FALSE)
+  if (is.numeric(res) && is.nan(res)) {
+    return(res)
   }
+
+  check_number_decimal(res, arg = I(paste0("Element at path `", xpath, "`")))
   res
 }
 
 #' @export
 xml_find_num.xml_nodeset <- function(x, xpath, ns = xml_ns(x)) {
-  if (length(x) == 0)
+  if (length(x) == 0) {
     return(numeric())
+  }
 
   vapply(x, function(x) xml_find_num(x, xpath = xpath, ns = ns), numeric(1))
 }
 
 #' @export
 xml_find_num.xml_missing <- function(x, xpath, ns = xml_ns(x)) {
-   numeric(0)
+  numeric(0)
+}
+
+#' @export
+#' @rdname xml_find_all
+xml_find_int <- function(x, xpath, ns = xml_ns(x)) {
+  UseMethod("xml_find_int")
+}
+
+#' @export
+xml_find_int.xml_node <- function(x, xpath, ns = xml_ns(x)) {
+  res <- .Call(xpath_search, x$node, x$doc, xpath, ns, Inf)
+  check_number_whole(res, arg = I(paste0("Element at path `", xpath, "`")))
+  as.integer(res)
+}
+
+#' @export
+xml_find_int.xml_nodeset <- function(x, xpath, ns = xml_ns(x)) {
+  if (length(x) == 0) {
+    return(integer())
+  }
+
+  vapply(x, function(x) xml_find_int(x, xpath = xpath, ns = ns), integer(1))
+}
+
+#' @export
+xml_find_int.xml_missing <- function(x, xpath, ns = xml_ns(x)) {
+  integer(0)
 }
 
 #' @export
@@ -161,23 +214,22 @@ xml_find_chr <- function(x, xpath, ns = xml_ns(x)) {
 #' @export
 xml_find_chr.xml_node <- function(x, xpath, ns = xml_ns(x)) {
   res <- .Call(xpath_search, x$node, x$doc, xpath, ns, Inf)
-  if (!is.character(res)) {
-    stop("result of type: ", sQuote(class(res)), ", not character", call. = FALSE)
-  }
+  check_string(res, arg = I(paste0("Element at path `", xpath, "`")))
   res
 }
 
 #' @export
 xml_find_chr.xml_nodeset <- function(x, xpath, ns = xml_ns(x)) {
-  if (length(x) == 0)
+  if (length(x) == 0) {
     return(character())
+  }
 
   vapply(x, function(x) xml_find_chr(x, xpath = xpath, ns = ns), character(1))
 }
 
 #' @export
 xml_find_chr.xml_missing <- function(x, xpath, ns = xml_ns(x)) {
-   character(0)
+  character(0)
 }
 
 #' @export
@@ -189,23 +241,22 @@ xml_find_lgl <- function(x, xpath, ns = xml_ns(x)) {
 #' @export
 xml_find_lgl.xml_node <- function(x, xpath, ns = xml_ns(x)) {
   res <- .Call(xpath_search, x$node, x$doc, xpath, ns, Inf)
-  if (!is.logical(res)) {
-    stop("result of type: ", sQuote(class(res)), ", not logical", call. = FALSE)
-  }
+  check_bool(res, arg = I(paste0("Element at path `", xpath, "`")))
   res
 }
 
 #' @export
 xml_find_lgl.xml_nodeset <- function(x, xpath, ns = xml_ns(x)) {
-  if (length(x) == 0)
+  if (length(x) == 0) {
     return(logical())
+  }
 
   vapply(x, function(x) xml_find_lgl(x, xpath = xpath, ns = ns), logical(1))
 }
 
 #' @export
 xml_find_lgl.xml_missing <- function(x, xpath, ns = xml_ns(x)) {
-   logical(0)
+  logical(0)
 }
 
 # Deprecated functions ----------------------------------------------------
