@@ -1,4 +1,8 @@
+#define R_NO_REMAP
 #include <Rinternals.h>
+#undef R_NO_REMAP
+
+#include <iterator>
 #include <vector>
 #include "xml2_utils.h"
 
